import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.metainfo.legacy import LegacyDefinition

from nomad.datamodel.metainfo import public
from nomad.datamodel.metainfo import common

m_package = Package(
    name='charmm_nomadmetainfo_json',
    description='None',
    a_legacy=LegacyDefinition(name='charmm.nomadmetainfo.json'))


class x_charmm_mdin_input_output_files(MCategory):
    '''
    Parameters of mdin belonging to x_charmm_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_charmm_mdin_input_output_files'))


class x_charmm_mdin_control_parameters(MCategory):
    '''
    Parameters of mdin belonging to x_charmm_section_control_parameters.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_charmm_mdin_control_parameters'))


class x_charmm_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_charmm_mdin_method'))


class x_charmm_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_charmm_mdout_single_configuration_calculation'))


class x_charmm_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category(
        a_legacy=LegacyDefinition(name='x_charmm_mdout_method'))


class x_charmm_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_charmm_mdout_run'))


class x_charmm_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category(
        categories=[public.settings_run],
        a_legacy=LegacyDefinition(name='x_charmm_mdin_run'))


class x_charmm_section_input_output_files(MSection):
    '''
    Section to store input and output file names
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_charmm_section_input_output_files'))


class x_charmm_section_control_parameters(MSection):
    '''
    Section to store the input and output control parameters
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_charmm_section_control_parameters'))

    x_charmm_inout_file_structure = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input topology file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_structure'))

    x_charmm_inout_file_trajectory = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output trajectory file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_trajectory'))

    x_charmm_inout_file_traj_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output trajectory file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_traj_coord'))

    x_charmm_inout_file_traj_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output file for velocities in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_traj_vel'))

    x_charmm_inout_file_traj_force = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output file for forces in the trajectory.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_traj_force'))

    x_charmm_inout_file_output_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_output_coord'))

    x_charmm_inout_file_out_coor_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output coordinates on log.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_out_coor_str'))

    x_charmm_inout_file_output_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_output_vel'))

    x_charmm_inout_file_output_force = Quantity(
        type=str,
        shape=[],
        description='''
        charmm output forces file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_output_force'))

    x_charmm_inout_file_input_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_input_coord'))

    x_charmm_inout_file_in_coor_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input coordinate on log file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_in_coor_str'))

    x_charmm_inout_file_input_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm input velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_input_vel'))

    x_charmm_inout_file_restart_coord = Quantity(
        type=str,
        shape=[],
        description='''
        charmm restart coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_restart_coord'))

    x_charmm_inout_file_restart_vel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm restart velocities file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_restart_vel'))

    x_charmm_inout_file_rtf_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm RTF residue file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_rtf_file'))

    x_charmm_inout_file_prm_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm PRM parameter file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_prm_file'))

    x_charmm_inout_file_cor_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm CRD coordinates file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_cor_file'))

    x_charmm_inout_file_stream = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream input/output.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_stream'))

    x_charmm_inout_file_rtf_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream RTF input.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_rtf_str'))

    x_charmm_inout_file_par_str = Quantity(
        type=str,
        shape=[],
        description='''
        charmm stream parameter input.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_par_str'))

    x_charmm_inout_file_output_log = Quantity(
        type=str,
        shape=[],
        description='''
        charmm MD output log file.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_file_output_log'))

    x_charmm_inout_control_gaussian_option_is = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_gaussian_option_is'))

    x_charmm_inout_control_crystal = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_crystal'))

    x_charmm_inout_control_crystal_type = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_crystal_type'))

    x_charmm_inout_control_a_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_a_length'))

    x_charmm_inout_control_b_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_b_length'))

    x_charmm_inout_control_c_length = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_c_length'))

    x_charmm_inout_control_alpha = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_alpha'))

    x_charmm_inout_control_beta = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_beta'))

    x_charmm_inout_control_gamma = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_gamma'))

    x_charmm_inout_control_mini = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_mini'))

    x_charmm_inout_control_nstep = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nstep'))

    x_charmm_inout_control_inbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_inbfrq'))

    x_charmm_inout_control_step = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_step'))

    x_charmm_inout_control_prtmin = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_prtmin'))

    x_charmm_inout_control_tolfun = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tolfun'))

    x_charmm_inout_control_tolgrd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tolgrd'))

    x_charmm_inout_control_tolitr = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tolitr'))

    x_charmm_inout_control_tolstp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tolstp'))

    x_charmm_inout_control_tfreq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tfreq'))

    x_charmm_inout_control_pcut = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pcut'))

    x_charmm_inout_control_ihbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ihbfrq'))

    x_charmm_inout_control_ncgcyc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ncgcyc'))

    x_charmm_inout_control_nprint = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nprint'))

    x_charmm_inout_control_nonbond_option_flags = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nonbond_option_flags'))

    x_charmm_inout_control_cutnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_cutnb'))

    x_charmm_inout_control_ctexnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ctexnb'))

    x_charmm_inout_control_ctonnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ctonnb'))

    x_charmm_inout_control_ctofnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ctofnb'))

    x_charmm_inout_control_cgonnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_cgonnb'))

    x_charmm_inout_control_cgofnb = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_cgofnb'))

    x_charmm_inout_control_wmin = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_wmin'))

    x_charmm_inout_control_cdie = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_cdie'))

    x_charmm_inout_control_switch = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_switch'))

    x_charmm_inout_control_vswitch = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_vswitch'))

    x_charmm_inout_control_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_atoms'))

    x_charmm_inout_control_wrnmxd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_wrnmxd'))

    x_charmm_inout_control_e14fac = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_e14fac'))

    x_charmm_inout_control_eps = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_eps'))

    x_charmm_inout_control_nbxmod = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nbxmod'))

    x_charmm_inout_control_hydrogen_bond_cutoff_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_cutoff_distance'))

    x_charmm_inout_control_hydrogen_bond_cutoff_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_cutoff_angle'))

    x_charmm_inout_control_hydrogen_bond_switching_on_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_switching_on_distance'))

    x_charmm_inout_control_hydrogen_bond_switching_off_distance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_switching_off_distance'))

    x_charmm_inout_control_hydrogen_bond_switching_on_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_switching_on_angle'))

    x_charmm_inout_control_hydrogen_bond_switching_off_angle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bond_switching_off_angle'))

    x_charmm_inout_control_hbond_exclusions_due_to_distance_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_exclusions_due_to_distance_cutoff'))

    x_charmm_inout_control_hbond_exclusions_due_to_angle_cutoff = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_exclusions_due_to_angle_cutoff'))

    x_charmm_inout_control_acceptor_antecedents = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_acceptor_antecedents'))

    x_charmm_inout_control_hbond_exclusions_due_to_duplications = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_exclusions_due_to_duplications'))

    x_charmm_inout_control_hbond_deletions_due_to_best_option = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_deletions_due_to_best_option'))

    x_charmm_inout_control_hbond_deletions_due_to_duplications = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_deletions_due_to_duplications'))

    x_charmm_inout_control_hbond_deletions_due_to_fixed_atoms = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_deletions_due_to_fixed_atoms'))

    x_charmm_inout_control_hbond_deletions_due_to_exclusion = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hbond_deletions_due_to_exclusion'))

    x_charmm_inout_control_hydrogen_bonds_present = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hydrogen_bonds_present'))

    x_charmm_inout_control_minimization_exit_status = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_minimization_exit_status'))

    x_charmm_inout_control_dyna = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_dyna'))

    x_charmm_inout_control_akmastp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_akmastp'))

    x_charmm_inout_control_firstt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_firstt'))

    x_charmm_inout_control_iseed = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iseed'))

    x_charmm_inout_control_iprfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iprfrq'))

    x_charmm_inout_control_ihtfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ihtfrq'))

    x_charmm_inout_control_ieqfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ieqfrq'))

    x_charmm_inout_control_iunrea = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iunrea'))

    x_charmm_inout_control_iunwri = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iunwri'))

    x_charmm_inout_control_iunos = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iunos'))

    x_charmm_inout_control_iuncrd = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iuncrd'))

    x_charmm_inout_control_iunvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iunvel'))

    x_charmm_inout_control_iunxyz = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iunxyz'))

    x_charmm_inout_control_kunit = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_kunit'))

    x_charmm_inout_control_nsavc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nsavc'))

    x_charmm_inout_control_nsavv = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nsavv'))

    x_charmm_inout_control_nsavx = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nsavx'))

    x_charmm_inout_control_iscale = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iscale'))

    x_charmm_inout_control_iscvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iscvel'))

    x_charmm_inout_control_iasors = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iasors'))

    x_charmm_inout_control_iasvel = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_iasvel'))

    x_charmm_inout_control_ichecw = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ichecw'))

    x_charmm_inout_control_ntrfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ntrfrq'))

    x_charmm_inout_control_ilbfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ilbfrq'))

    x_charmm_inout_control_imgfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_imgfrq'))

    x_charmm_inout_control_isvfrq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_isvfrq'))

    x_charmm_inout_control_ncycle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_ncycle'))

    x_charmm_inout_control_nsnos = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_nsnos'))

    x_charmm_inout_control_teminc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_teminc'))

    x_charmm_inout_control_tstruc = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tstruc'))

    x_charmm_inout_control_finalt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_finalt'))

    x_charmm_inout_control_twindh = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_twindh'))

    x_charmm_inout_control_twindl = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_twindl'))

    x_charmm_inout_control_time_step = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_time_step'))

    x_charmm_inout_control_random_num_gen_seeds = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_random_num_gen_seeds'))

    x_charmm_inout_control_number_of_degrees_of_freedom = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_number_of_degrees_of_freedom'))

    x_charmm_inout_control_gaussian_option = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_gaussian_option'))

    x_charmm_inout_control_velocities_assigned_at_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_velocities_assigned_at_temperature'))

    x_charmm_inout_control_seeds = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_seeds'))

    x_charmm_inout_control_shake_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_shake_tolerance'))

    x_charmm_inout_control_strt = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_strt'))

    x_charmm_inout_control_rest = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_rest'))

    x_charmm_inout_control_qref = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_qref'))

    x_charmm_inout_control_tref = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tref'))

    x_charmm_inout_control_hoover = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_hoover'))

    x_charmm_inout_control_reft = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_reft'))

    x_charmm_inout_control_tmass = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_tmass'))

    x_charmm_inout_control_pcons = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pcons'))

    x_charmm_inout_control_pmass = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pmass'))

    x_charmm_inout_control_number_of_steps = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_number_of_steps'))

    x_charmm_inout_control_steps_per_cycle = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_steps_per_cycle'))

    x_charmm_inout_control_initial_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_initial_temperature'))

    x_charmm_inout_control_dielectric = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_dielectric'))

    x_charmm_inout_control_excluded_species_or_groups = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_excluded_species_or_groups'))

    x_charmm_inout_control_1_4_electrostatics_scale = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_1_4_electrostatics_scale'))

    x_charmm_inout_control_velocity_rescale_temp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_velocity_rescale_temp'))

    x_charmm_inout_control_velocity_reassignment_freq = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_velocity_reassignment_freq'))

    x_charmm_inout_control_velocity_reassignment_temp = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_velocity_reassignment_temp'))

    x_charmm_inout_control_langevin_dynamics = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_dynamics'))

    x_charmm_inout_control_langevin_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_temperature'))

    x_charmm_inout_control_langevin_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_integrator'))

    x_charmm_inout_control_langevin_damping_coefficient_unit = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_damping_coefficient_unit'))

    x_charmm_inout_control_temperature_coupling = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_temperature_coupling'))

    x_charmm_inout_control_coupling_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_coupling_temperature'))

    x_charmm_inout_control_target_pressure = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_target_pressure'))

    x_charmm_inout_control_langevin_oscillation_period = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_oscillation_period'))

    x_charmm_inout_control_langevin_decay_time = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_decay_time'))

    x_charmm_inout_control_langevin_piston_temperature = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_langevin_piston_temperature'))

    x_charmm_inout_control_pressure_control = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pressure_control'))

    x_charmm_inout_control_particle_mesh_ewald = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_particle_mesh_ewald'))

    x_charmm_inout_control_pme_tolerance = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pme_tolerance'))

    x_charmm_inout_control_pme_ewald_coefficient = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pme_ewald_coefficient'))

    x_charmm_inout_control_pme_interpolation_order = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pme_interpolation_order'))

    x_charmm_inout_control_pme_grid_dimensions = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pme_grid_dimensions'))

    x_charmm_inout_control_pme_maximum_grid_spacing = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_pme_maximum_grid_spacing'))

    x_charmm_inout_control_minimization = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_minimization'))

    x_charmm_inout_control_verlet_integrator = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_verlet_integrator'))

    x_charmm_inout_control_random_number_seed = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_random_number_seed'))

    x_charmm_inout_control_structure_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_structure_file'))

    x_charmm_inout_control_parameter_file = Quantity(
        type=str,
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_parameter_file'))

    x_charmm_inout_control_number_of_parameters = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_number_of_parameters'))

    x_charmm_inout_control_parameters = Quantity(
        type=str,
        shape=['x_charmm_inout_control_number_of_parameters'],
        description='''
        charmm running environment and control parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_inout_control_parameters'))

    x_charmm_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_charmm_section_input_output_files'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_charmm_section_input_output_files'))


class x_charmm_section_atom_to_atom_type_ref(MSection):
    '''
    Section to store atom label to atom type definition list
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_charmm_section_atom_to_atom_type_ref'))

    x_charmm_atom_to_atom_type_ref = Quantity(
        type=np.dtype(np.int64),
        shape=['number_of_atoms_per_type'],
        description='''
        Reference to the atoms of each atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_to_atom_type_ref'))


class x_charmm_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False, a_legacy=LegacyDefinition(name='x_charmm_section_single_configuration_calculation'))


class section_system(public.section_system):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_system'))

    x_charmm_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_positions_image_index'))

    x_charmm_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_positions_scaled'))

    x_charmm_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_positions_wrapped'))

    x_charmm_lattice_lengths = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Lattice dimensions in a vector. Vector includes [a, b, c] lengths.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_charmm_lattice_lengths'))

    x_charmm_lattice_angles = Quantity(
        type=np.dtype(np.float64),
        shape=[3],
        description='''
        Angles of lattice vectors. Vector includes [alpha, beta, gamma] in degrees.
        ''',
        categories=[public.configuration_core],
        a_legacy=LegacyDefinition(name='x_charmm_lattice_angles'))

    x_charmm_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_dummy'))

    x_charmm_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_mdin_finline'))

    x_charmm_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_timestep_store'))

    x_charmm_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_number_of_atoms_store'))

    x_charmm_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_box_bound_store'))

    x_charmm_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_box_bounds_store'))

    x_charmm_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_variables_store'))

    x_charmm_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_traj_atoms_store'))


class section_sampling_method(public.section_sampling_method):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_sampling_method'))

    x_charmm_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''',
        categories=[public.settings_barostat, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_barostat_target_pressure'))

    x_charmm_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''',
        categories=[public.settings_barostat, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_barostat_tau'))

    x_charmm_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''',
        categories=[public.settings_barostat, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_barostat_type'))

    x_charmm_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''',
        categories=[public.settings_integrator, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_integrator_dt'))

    x_charmm_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''',
        categories=[public.settings_integrator, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_integrator_type'))

    x_charmm_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''',
        categories=[public.settings_integrator, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_periodicity_type'))

    x_charmm_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_charmm_langevin_gamma'))

    x_charmm_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''',
        categories=[public.settings_integrator, public.settings_sampling, public.settings_molecular_dynamics],
        a_legacy=LegacyDefinition(name='x_charmm_number_of_steps_requested'))

    x_charmm_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_charmm_thermostat_level'))

    x_charmm_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_charmm_thermostat_target_temperature'))

    x_charmm_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_charmm_thermostat_tau'))

    x_charmm_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''',
        categories=[public.settings_sampling, public.settings_molecular_dynamics, public.settings_thermostat],
        a_legacy=LegacyDefinition(name='x_charmm_thermostat_type'))


class section_atom_type(common.section_atom_type):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_atom_type'))

    x_charmm_atom_name = Quantity(
        type=str,
        shape=[],
        description='''
        Atom name of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_name'))

    x_charmm_atom_type = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_type'))

    x_charmm_atom_element = Quantity(
        type=str,
        shape=[],
        description='''
        Atom type of an atom in topology definition.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_element'))

    x_charmm_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_type_element'))

    x_charmm_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_atom_type_radius'))

    number_of_atoms_per_type = Quantity(
        type=int,
        shape=[],
        description='''
        Number of atoms involved in this type.
        ''',
        a_legacy=LegacyDefinition(name='number_of_atoms_per_type'))


class section_interaction(common.section_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_interaction'))

    x_charmm_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_interaction_atom_to_atom_type_ref'))

    x_charmm_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_defined_pair_interactions'))

    x_charmm_pair_interaction_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_charmm_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_pair_interaction_atom_type_ref'))

    x_charmm_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_pair_interaction_parameters'))


class section_molecule_interaction(common.section_molecule_interaction):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_molecule_interaction'))

    x_charmm_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_molecule_interaction_atom_to_atom_type_ref'))

    x_charmm_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_defined_molecule_pair_interactions'))

    x_charmm_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_pair_molecule_interaction_parameters'))

    x_charmm_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=common.section_atom_type,
        shape=['x_charmm_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_pair_molecule_interaction_to_atom_type_ref'))


class section_run(public.section_run):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_run'))

    x_charmm_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_version_date'))

    x_charmm_parallel_task_nr = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Program task no.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_parallel_task_nr'))

    x_charmm_build_osarch = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build OS/ARCH
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_build_osarch'))

    x_charmm_output_created_by_user = Quantity(
        type=str,
        shape=[],
        description='''
        Output file creator
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_output_created_by_user'))

    x_charmm_most_severe_warning_level = Quantity(
        type=str,
        shape=[],
        description='''
        Highest charmm warning level in the run.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_most_severe_warning_level'))

    x_charmm_program_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program Build date
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_build_date'))

    x_charmm_program_citation = Quantity(
        type=str,
        shape=[],
        description='''
        Program citations
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_citation'))

    x_charmm_program_copyright = Quantity(
        type=str,
        shape=[],
        description='''
        Program copyright
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_copyright'))

    x_charmm_number_of_tasks = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of tasks in parallel program (MPI).
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_tasks'))

    x_charmm_program_module_version = Quantity(
        type=str,
        shape=[],
        description='''
        charmm program module version.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_module_version'))

    x_charmm_program_license = Quantity(
        type=str,
        shape=[],
        description='''
        charmm program license.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_license'))

    x_charmm_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_xlo_xhi'))

    x_charmm_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_file_store'))

    x_charmm_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_working_path'))

    x_charmm_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_execution_host'))

    x_charmm_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_execution_path'))

    x_charmm_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_module'))

    x_charmm_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_execution_date'))

    x_charmm_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_program_execution_time'))

    x_charmm_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_mdin_header'))

    x_charmm_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_mdin_wt'))

    x_charmm_section_control_parameters = SubSection(
        sub_section=SectionProxy('x_charmm_section_control_parameters'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_charmm_section_control_parameters'))


class section_topology(common.section_topology):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_topology'))

    x_charmm_input_units_store = Quantity(
        type=str,
        shape=[],
        description='''
        It determines the units of all quantities specified in the input script and data
        file, as well as quantities output to the screen, log file, and dump files.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_input_units_store'))

    x_charmm_data_bond_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_bond_types_store'))

    x_charmm_data_bond_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_bond_count_store'))

    x_charmm_data_angle_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_angle_count_store'))

    x_charmm_data_atom_types_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_atom_types_store'))

    x_charmm_data_dihedral_count_store = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_dihedral_count_store'))

    x_charmm_data_angles_store = Quantity(
        type=str,
        shape=[],
        description='''
        store temporarly
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_angles_store'))

    x_charmm_data_angle_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_angle_list_store'))

    x_charmm_data_bond_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_bond_list_store'))

    x_charmm_data_dihedral_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_dihedral_list_store'))

    x_charmm_data_dihedral_coeff_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_dihedral_coeff_list_store'))

    x_charmm_masses_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_masses_store'))

    x_charmm_data_topo_list_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_data_topo_list_store'))

    x_charmm_section_atom_to_atom_type_ref = SubSection(
        sub_section=SectionProxy('x_charmm_section_atom_to_atom_type_ref'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_charmm_section_atom_to_atom_type_ref'))


class section_frame_sequence(public.section_frame_sequence):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_frame_sequence'))

    x_charmm_number_of_volumes_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of volumes in this sequence of frames, see
        x_charmm_frame_sequence_volume.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_volumes_in_sequence'))

    x_charmm_number_of_densities_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of densities in this sequence of frames, see
        x_charmm_frame_sequence_density.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_densities_in_sequence'))

    x_charmm_number_of_bond_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_bond_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_bond_energies_in_sequence'))

    x_charmm_number_of_urey_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_urey_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_urey_energies_in_sequence'))

    x_charmm_number_of_high_frequency_correction_total_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_high_frequency_correction_total_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_high_frequency_correction_total_energies_in_sequence'))

    x_charmm_number_of_high_frequency_correction_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_high_frequency_correction_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_high_frequency_correction_energies_in_sequence'))

    x_charmm_number_of_high_frequency_correction_kinetic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_high_frequency_correction_kinetic_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_high_frequency_correction_kinetic_energies_in_sequence'))

    x_charmm_number_of_virial_kinetic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of bond_energies in this sequence of frames, see
        x_charmm_frame_sequence_virial_kinetic_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_virial_kinetic_energies_in_sequence'))

    x_charmm_number_of_angle_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of angle_energies in this sequence of frames, see
        x_charmm_frame_sequence_angle_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_angle_energies_in_sequence'))

    x_charmm_number_of_proper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of proper_dihedral_energies in this sequence of frames, see
        x_charmm_frame_sequence_proper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_proper_dihedral_energies_in_sequence'))

    x_charmm_number_of_improper_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of improper_dihedral_energies in this sequence of frames, see
        x_charmm_frame_sequence_improper_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_improper_dihedral_energies_in_sequence'))

    x_charmm_number_of_cmap_dihedral_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of cmap_dihedral_energies in this sequence of frames, see
        x_charmm_frame_sequence_cmap_dihedral_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_cmap_dihedral_energies_in_sequence'))

    x_charmm_number_of_vdw_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of vdw_energies in this sequence of frames, see
        x_charmm_frame_sequence_vdw_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_vdw_energies_in_sequence'))

    x_charmm_number_of_boundary_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of boundary_energies in this sequence of frames, see
        x_charmm_frame_sequence_boundary_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_boundary_energies_in_sequence'))

    x_charmm_number_of_electrostatic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of electrostatic_energies in this sequence of frames, see
        x_charmm_frame_sequence_electrostatic_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_electrostatic_energies_in_sequence'))

    x_charmm_number_of_total_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total_energies in this sequence of frames, see
        x_charmm_frame_sequence_total_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_total_energies_in_sequence'))

    x_charmm_number_of_total_kinetic_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of total_kinetic_energies in this sequence of frames, see
        x_charmm_frame_sequence_total_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_total_kinetic_energies_in_sequence'))

    x_charmm_number_of_misc_energies_in_sequence = Quantity(
        type=int,
        shape=[],
        description='''
        Gives the number of misc_energies in this sequence of frames, see
        x_charmm_frame_sequence_misc_energy.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_number_of_misc_energies_in_sequence'))

    x_charmm_frame_sequence_density_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_densities_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_densities values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_density_frames'))

    x_charmm_frame_sequence_density = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_densities_in_sequence'],
        description='''
        Array containing the values of the density along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_density_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_density'))

    x_charmm_frame_sequence_cmap_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_cmap_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_cmap_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_cmap_dihedral_energy_frames'))

    x_charmm_frame_sequence_cmap_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_cmap_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the cmap_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_cmap_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_cmap_dihedral_energy'))

    x_charmm_frame_sequence_improper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_improper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_improper_dihedral_energy_frames'))

    x_charmm_frame_sequence_improper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_improper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the improper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_improper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_improper_dihedral_energy'))

    x_charmm_frame_sequence_proper_dihedral_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_proper_dihedral_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_proper_dihedral_energy_frames'))

    x_charmm_frame_sequence_proper_dihedral_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_proper_dihedral_energy_in_sequence'],
        description='''
        Array containing the values of the proper_dihedral_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_proper_dihedral_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_proper_dihedral_energy'))

    x_charmm_frame_sequence_bond_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_bond_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_bond_energy_frames'))

    x_charmm_frame_sequence_urey_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_urey_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_urey_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_urey_energy_frames'))

    x_charmm_frame_sequence_high_frequency_correction_total_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_high_frequency_correction_total_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_high_frequency_correction_total_energy values refers to.
        If not given it defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_total_energy_frames'))

    x_charmm_frame_sequence_high_frequency_correction_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_high_frequency_correction_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_high_frequency_correction_energy values refers to. If not
        given it defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_energy_frames'))

    x_charmm_frame_sequence_high_frequency_correction_kinetic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_high_frequency_correction_kinetic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_high_frequency_correction_kinetic_energy values refers to.
        If not given it defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_kinetic_energy_frames'))

    x_charmm_frame_sequence_virial_kinetic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_virial_kinetic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_virial_kinetic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_virial_kinetic_energy_frames'))

    x_charmm_frame_sequence_bond_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_bond_energies_in_sequence'],
        description='''
        Array containing the values of the bond_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_bond_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_bond_energy'))

    x_charmm_frame_sequence_urey_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_urey_energies_in_sequence'],
        description='''
        Array containing the values of the urey_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_urey_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_urey_energy'))

    x_charmm_frame_sequence_high_frequency_correction_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_high_frequency_correction_total_energies_in_sequence'],
        description='''
        Array containing the values of the high_frequency_correction_total_energy along
        this sequence of frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_high_frequency_correction_total_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_total_energy'))

    x_charmm_frame_sequence_high_frequency_correction_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_high_frequency_correction_energies_in_sequence'],
        description='''
        Array containing the values of the high_frequency_correction_energy along this
        sequence of frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_high_frequency_correction_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_energy'))

    x_charmm_frame_sequence_high_frequency_correction_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_high_frequency_correction_kinetic_energies_in_sequence'],
        description='''
        Array containing the values of the high_frequency_correction_kinetic_energy along
        this sequence of frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_high_frequency_correction_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_high_frequency_correction_kinetic_energy'))

    x_charmm_frame_sequence_virial_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_virial_kinetic_energies_in_sequence'],
        description='''
        Array containing the values of the virial_kinetic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_virial_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_virial_kinetic_energy'))

    x_charmm_frame_sequence_boundary_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_boundary values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_boundary_frames'))

    x_charmm_frame_sequence_boundary = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_boundary_energies_in_sequence'],
        description='''
        Array containing the values of the boundary along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_boundary_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_boundary'))

    x_charmm_frame_sequence_angle_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_angle_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_angle_energy_frames'))

    x_charmm_frame_sequence_angle_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_angle_energies_in_sequence'],
        description='''
        Array containing the values of the angle_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_angle_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_angle_energy'))

    x_charmm_frame_sequence_vdw_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_vdw_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_vdw_energy_frames'))

    x_charmm_frame_sequence_vdw_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_vdw_energies_in_sequence'],
        description='''
        Array containing the values of the vdw_energy along this sequence of frames (i.e.,
        a trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_vdw_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_vdw_energy'))

    x_charmm_frame_sequence_electrostatic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_electrostatic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_electrostatic_energy_frames'))

    x_charmm_frame_sequence_electrostatic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_electrostatic_energies_in_sequence'],
        description='''
        Array containing the values of the electrostatic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_electrostatic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_electrostatic_energy'))

    x_charmm_frame_sequence_total_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_total_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_total_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_total_energy_frames'))

    x_charmm_frame_sequence_total_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_total_energies_in_sequence'],
        description='''
        Array containing the values of the total_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_total_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_total_energy'))

    x_charmm_frame_sequence_total_kinetic_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_total_kinetic_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_total_kinetic_energy values refers to. If not given it
        defaults to the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_total_kinetic_energy_frames'))

    x_charmm_frame_sequence_total_kinetic_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_total_kinetic_energies_in_sequence'],
        description='''
        Array containing the values of the total_kinetic_energy along this sequence of
        frames (i.e., a trajectory, a frame is one
        section_single_configuration_calculation). If not all frames have a value the
        indices of the frames that have a value are stored in
        frame_sequence_total_kinetic_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_total_kinetic_energy'))

    x_charmm_frame_sequence_misc_energy_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_misc_energy values refers to. If not given it defaults to
        the trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_misc_energy_frames'))

    x_charmm_frame_sequence_misc_energy = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_misc_energies_in_sequence'],
        description='''
        Array containing the values of the misc_energy along this sequence of frames
        (i.e., a trajectory, a frame is one section_single_configuration_calculation). If
        not all frames have a value the indices of the frames that have a value are stored
        in frame_sequence_misc_energy_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_misc_energy'))

    x_charmm_frame_sequence_volume_frames = Quantity(
        type=np.dtype(np.int32),
        shape=['x_charmm_number_of_volumes_in_sequence'],
        description='''
        Array containing the strictly increasing indices of the frames the
        x_charmm_frame_sequence_volume values refers to. If not given it defaults to the
        trivial mapping 0,1,...
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_volume_frames'))

    x_charmm_frame_sequence_volume = Quantity(
        type=np.dtype(np.float64),
        shape=['x_charmm_number_of_volumes_in_sequence'],
        description='''
        Array containing the values of the volume along this sequence of frames (i.e., a
        trajectory, a frame is one section_single_configuration_calculation). If not all
        frames have a value the indices of the frames that have a value are stored in
        frame_sequence_volume_frames.
        ''',
        a_legacy=LegacyDefinition(name='x_charmm_frame_sequence_volume'))


class section_single_configuration_calculation(public.section_single_configuration_calculation):

    m_def = Section(validate=False, extends_base_section=True, a_legacy=LegacyDefinition(name='section_single_configuration_calculation'))

    x_charmm_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_charmm_section_single_configuration_calculation'),
        repeats=True,
        a_legacy=LegacyDefinition(name='x_charmm_section_single_configuration_calculation'))


m_package.__init_metainfo__()



import re
from typing import Tuple
from pathlib import Path

from invoke import task

from optimade import __version__

TOP_DIR = Path(__file__).parent.resolve()


def update_file(filename: str, sub_line: Tuple[str, str], strip: str = None):
    """Utility function for tasks to read, update, and write files"""
    with open(filename, "r") as handle:
        lines = [
            re.sub(sub_line[0], sub_line[1], line.rstrip(strip)) for line in handle
        ]

    with open(filename, "w") as handle:
        handle.write("\n".join(lines))
        handle.write("\n")


@task
def setver(_, patch=False, new_ver=""):
    if (not patch and not new_ver) or (patch and new_ver):
        raise Exception(
            "Either use --patch or specify e.g. --new-ver='Major.Minor.Patch(a|b|rc)?[0-9]+'"
        )
    if patch:
        v = [int(x) for x in __version__.split(".")]
        v[2] += 1
        new_ver = ".".join(map(str, v))

    update_file(
        TOP_DIR.joinpath("optimade/__init__.py"),
        ("__version__ = .+", f'__version__ = "{new_ver}"'),
    )
    update_file(
        TOP_DIR.joinpath("setup.py"), ("version=([^,]+),", f'version="{new_ver}",')
    )

    print("Bumped version to {}".format(new_ver))


@task
def update_openapijson(c):
    # pylint: disable=import-outside-toplevel
    from optimade.server.main import app, update_schema
    from optimade.server.main_index import (
        app as app_index,
        update_schema as update_schema_index,
    )

    update_schema(app)
    update_schema_index(app_index)

    c.run(
        f"cp {TOP_DIR.joinpath('openapi/local_openapi.json')} {TOP_DIR.joinpath('openapi/openapi.json')}"
    )
    c.run(
        f"cp {TOP_DIR.joinpath('openapi/local_index_openapi.json')} {TOP_DIR.joinpath('openapi/index_openapi.json')}"
    )


@task
def set_optimade_ver(_, ver=""):
    if not ver:
        raise Exception("Please specify --ver='Major.Minor.Patch'")
    if not re.match("[0-9]+.[0-9]+.[0-9]+", ver):
        raise Exception("ver MUST be expressed as 'Major.Minor.Patch'")

    update_file(
        TOP_DIR.joinpath("optimade/__init__.py"),
        ("__api_version__ = .+", f'__api_version__ = "{ver}"'),
    )
    update_file(
        TOP_DIR.joinpath(".ci/optimade-version.json"),
        ('"message": .+', f'"message": "v{ver}",'),
    )
    for regex, version in (
        ("[0-9]+", ver.split(".")[0]),
        ("[0-9]+.[0-9]+", ".".join(ver.split(".")[:2])),
        ("[0-9]+.[0-9]+.[0-9]+", ver),
    ):
        update_file(
            TOP_DIR.joinpath("README.md"),
            (f"example/v{regex}", f"example/v{version}"),
            strip="\n",
        )
    update_file(
        TOP_DIR.joinpath(".github/workflows/validator_action.yml"),
        ("/v[0-9]+", f"/v{ver.split('.')[0]}"),
    )
    update_file(
        TOP_DIR.joinpath("README.md"), ("v[0-9]+", f"v{ver.split('.')[0]}"), strip="\n"
    )
    update_file(TOP_DIR.joinpath("action.yml"), ("/v[0-9]+", f"/v{ver.split('.')[0]}"))
    update_file(
        TOP_DIR.joinpath("optimade/validator/github_action/entrypoint.sh"),
        (
            "'[0-9]+' '[0-9]+.[0-9]+' '[0-9]+.[0-9]+.[0-9]+'",
            f"'{ver.split('.')[0]}' '{'.'.join(ver.split('.')[:2])}' '{ver}'",
        ),
    )
    update_file(
        TOP_DIR.joinpath("INSTALL.md"), (r"/v[0-9]+(\.[0-9]+){2}", f"/v{version}")
    )

    print(f"Bumped OPTIMADE version to {ver}")


@task
def parse_spec_for_filters(_):
    import requests

    filter_path = TOP_DIR.joinpath("optimade/validator/data/filters.txt")
    optional_filter_path = TOP_DIR.joinpath(
        "optimade/validator/data/optional_filters.txt"
    )

    specification_flines = (
        requests.get(
            "https://raw.githubusercontent.com/Materials-Consortia/OPTIMADE/develop/optimade.rst"
        )
        .content.decode("utf-8")
        .split("\n")
    )

    filters = []
    optional_filters = []
    optional_triggers = ("OPTIONAL",)
    for line in specification_flines:
        if ":filter:" in line:
            for _split in line.replace("filter=", "").split(":filter:")[1:]:
                _filter = _split.split("`")[1].strip()
                if any(trigger in line for trigger in optional_triggers):
                    optional_filters.append(_filter)
                else:
                    filters.append(_filter)

    with open(filter_path, "w") as f:
        for _filter in filters:
            f.write(_filter + "\n")

    with open(optional_filter_path, "w") as f:
        for _filter in optional_filters:
            f.write(_filter + "\n")



""" This module contains a validator class that can be pointed
at an OPTIMADE implementation and validated against the pydantic
models in this package.

"""
# pylint: disable=import-outside-toplevel

import time
import requests
import sys
import logging
import urllib.parse

try:
    import simplejson as json
except ImportError:
    import json

from typing import Union

from pydantic import ValidationError
from fastapi.testclient import TestClient

from optimade.models import InfoResponse, EntryInfoResponse, IndexInfoResponse

from .data import MANDATORY_FILTER_EXAMPLES, OPTIONAL_FILTER_EXAMPLES
from .validator_model_patches import (
    ValidatorLinksResponse,
    ValidatorEntryResponseOne,
    ValidatorEntryResponseMany,
    ValidatorReferenceResponseOne,
    ValidatorReferenceResponseMany,
    ValidatorStructureResponseOne,
    ValidatorStructureResponseMany,
)


BASE_INFO_ENDPOINT = "info"
LINKS_ENDPOINT = "links"
REQUIRED_ENTRY_ENDPOINTS = ["references", "structures"]

ENDPOINT_MANDATORY_QUERIES = {
    "structures": MANDATORY_FILTER_EXAMPLES,
    "references": [],
}

ENDPOINT_OPTIONAL_QUERIES = {
    "structures": OPTIONAL_FILTER_EXAMPLES,
    "references": [],
}

RESPONSE_CLASSES = {
    "references": ValidatorReferenceResponseMany,
    "references/": ValidatorReferenceResponseOne,
    "structures": ValidatorStructureResponseMany,
    "structures/": ValidatorStructureResponseOne,
    "info": InfoResponse,
    "links": ValidatorLinksResponse,
}
RESPONSE_CLASSES.update(
    {f"info/{entry}": EntryInfoResponse for entry in REQUIRED_ENTRY_ENDPOINTS}
)

REQUIRED_ENTRY_ENDPOINTS_INDEX = []
RESPONSE_CLASSES_INDEX = {"info": IndexInfoResponse, "links": ValidatorLinksResponse}


def print_warning(string, **kwargs):
    """ Print but angry. """
    print(f"\033[93m{string}\033[0m", **kwargs)


def print_failure(string, **kwargs):
    """ Print but sad. """
    print(f"\033[91m\033[4m{string}\033[0m", **kwargs)


def print_success(string, **kwargs):
    """ Print but happy. """
    print(f"\033[92m\033[1m{string}\033[0m", **kwargs)


class ResponseError(Exception):
    """ This exception should be raised for a manual hardcoded test failure. """


class Client:  # pragma: no cover
    def __init__(self, base_url: str, max_retries=5):
        """ Initialises the Client with the given `base_url` without testing
        if it is valid.

        Parameters:
            base_url (str): the base URL of the optimade implementation, including
            request protocol (e.g. `'http://'`) and API version number if necessary.
            Examples:
                - `'http://example.org/optimade/v1'`,
                - `'www.crystallography.net/cod-test/optimade/v0.10.0/'`
            Note: A maximum of one slash ("/") is allowed as the last character.

        """
        self.base_url = base_url
        self.last_request = None
        self.response = None
        self.max_retries = max_retries

    def get(self, request: str):
        """ Makes the given request, with a number of retries if being rate limited. The
        request will be prepended with the `base_url` unless the request appears to be an
        absolute URL (i.e. starts with `http://` or `https://`).

        Parameters:
            request (str): the request to make against the base URL of this client.

        Returns:
            response (requests.models.Response): the response from the server.

        Raises:
            SystemExit: if there is no response from the server, or if the URL is invalid.
            ResponseError: if the server does not respond with a non-429 status code within
                the `MAX_RETRIES` attempts.

        """
        if urllib.parse.urlparse(request, allow_fragments=True).scheme:
            self.last_request = request
        else:
            if request and not request.startswith("/"):
                request = f"/{request}"
            self.last_request = f"{self.base_url}{request}"

        status_code = None
        retries = 0
        # probably a smarter way to do this with requests, but their documentation 404's...
        while retries < self.max_retries:
            retries += 1
            try:
                self.response = requests.get(self.last_request)
            except requests.exceptions.ConnectionError:
                sys.exit(f"No response from server at {self.last_request}")
            except requests.exceptions.MissingSchema:
                sys.exit(
                    f"Unable to make request on {self.last_request}, did you mean http://{self.last_request}?"
                )
            status_code = self.response.status_code
            if status_code != 429:
                break

            print("Hit rate limit, sleeping for 1 s...")
            time.sleep(1)

        else:
            raise ResponseError("Hit max (manual) retries on request.")

        return self.response


def test_case(test_fn):
    """ Wrapper for test case functions, which pretty_prints any errors
    depending on verbosity level and returns only the response to the caller.

    Parameters:
        test_fn (callable): function that returns a response to pass to caller,
            and a message to print upon success. Should raise `ResponseError`,
            `ValidationError` or `ManualValidationError` if the test case has failed.

    Keyword arguments:
        optional (bool): whether or not to treat the test as optional.

    """
    from functools import wraps

    @wraps(test_fn)
    def wrapper(validator, *args, optional=False, **kwargs):
        try:
            result, msg = test_fn(validator, *args, **kwargs)
        except json.JSONDecodeError as exc:
            result = None
            msg = (
                "Critical: unable to parse server response as JSON. "
                f"Error: {type(exc).__name__}: {exc}"
            )
        except (ResponseError, ValidationError) as exc:
            result = None
            msg = f"{type(exc).__name__}: {exc}"

        try:
            request = validator.client.last_request
        except AttributeError:
            request = validator.base_url

        if result is not None:
            if not optional:
                validator.success_count += 1
            else:
                validator.optional_success_count += 1
            message = f"✔: {request} - {msg}"
            if validator.verbosity > 0:
                if optional:
                    print(message)
                else:
                    print_success(message)
            else:
                if optional:
                    print(".", end="", flush=True)
                else:
                    print_success(".", end="", flush=True)
        else:
            if not optional:
                validator.failure_count += 1
            else:
                validator.optional_failure_count += 1
            request = request.replace("\n", "")
            message = f"{msg}".split("\n")
            summary = f"✖: {request} - {test_fn.__name__} - failed with error"
            validator.failure_messages.append((summary, message))
            if validator.verbosity > 0:
                if optional:
                    print(summary)
                    for line in message:
                        print(f"\t{line}")
                else:
                    print_failure(summary)
                    for line in message:
                        print_warning(f"\t{line}")
            else:
                if optional:
                    print("✖", end="", flush=True)
                else:
                    print_failure("✖", end="", flush=True)

        return result

    return wrapper


class ImplementationValidator:
    """ Class to call test functions on a particular OPTIMADE
    implementation.

    Uses the pydantic models in `optimade.models` to validate the
    response from the server and crawl through the available endpoints.

    Drawbacks:
        - only works for current version of the specification as defined
        by `optimade.models`.

    """

    def __init__(  # pylint: disable=too-many-arguments
        self,
        client: Union[Client, TestClient] = None,
        base_url: str = None,
        verbosity: int = 0,
        page_limit: int = 5,
        max_retries: int = 5,
        as_type: str = None,
        index: bool = False,
    ):
        """ Set up the tests to run, based on constants in this module
        for required endpoints.

        """

        self.verbosity = verbosity
        self.max_retries = max_retries
        self.page_limit = page_limit
        self.index = index

        if as_type is None:
            self.as_type_cls = None
        elif self.index:
            if as_type not in RESPONSE_CLASSES_INDEX.keys():
                raise RuntimeError(
                    f"Provided as_type='{as_type}' not allowed for an Index meta-database."
                )
            self.as_type_cls = RESPONSE_CLASSES_INDEX[as_type]
        elif as_type in ("structure", "reference"):
            self.as_type_cls = RESPONSE_CLASSES[f"{as_type}s/"]
        else:
            self.as_type_cls = RESPONSE_CLASSES[as_type]

        if client is None and base_url is None:
            raise RuntimeError(
                "Need at least a URL or a client to initialize validator."
            )
        if base_url and client:
            raise RuntimeError("Please specify at most one of base_url or client.")
        if client:
            self.client = client
            self.base_url = self.client.base_url
        else:
            while base_url.endswith("/"):
                base_url = base_url[:-1]
            self.base_url = base_url
            self.client = Client(base_url, max_retries=self.max_retries)

        self.test_id_by_type = {}
        self._setup_log()
        self.expected_entry_endpoints = (
            REQUIRED_ENTRY_ENDPOINTS_INDEX if self.index else REQUIRED_ENTRY_ENDPOINTS
        )
        self.test_entry_endpoints = set(self.expected_entry_endpoints)
        self.endpoint_mandatory_queries = (
            {} if self.index else ENDPOINT_MANDATORY_QUERIES
        )

        self.endpoint_optional_queries = {} if self.index else ENDPOINT_OPTIONAL_QUERIES

        self.response_classes = (
            RESPONSE_CLASSES_INDEX if self.index else RESPONSE_CLASSES
        )

        # some simple checks on base_url
        base_url = urllib.parse.urlparse(self.base_url)
        # only allow filters/endpoints if we are working in "as_type" mode
        if self.as_type_cls is None and (
            base_url.query
            or any(endp in base_url.path for endp in self.expected_entry_endpoints)
        ):
            raise SystemExit(
                "Base URL not appropriate: should not contain an endpoint or filter."
            )

        # if valid is True on exit, script returns 0 to shell
        # if valid is False on exit, script returns 1 to shell
        # if valid is None on exit, script returns 2 to shell, indicating an internal failure
        self.valid = None

        self.success_count = 0
        self.failure_count = 0
        self.optional_success_count = 0
        self.optional_failure_count = 0
        self.failure_messages = []

    def _setup_log(self):
        """ Define stdout log based on given verbosity. """
        self._log = logging.getLogger(__name__)
        self._log.handlers = []
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(
            logging.Formatter("%(asctime)s - %(name)s | %(levelname)8s: %(message)s")
        )
        self._log.addHandler(stdout_handler)
        if self.verbosity == 0:
            self._log.setLevel(logging.CRITICAL)
        elif self.verbosity == 1:
            self._log.setLevel(logging.INFO)
        else:
            self._log.setLevel(logging.DEBUG)

    def main(self):
        """ Run all the test cases of the implementation, or the single type test. """

        # if single type has been set, only run that test
        if self.as_type_cls is not None:
            self._log.info(
                "Validating response of %s with model %s",
                self.base_url,
                self.as_type_cls,
            )
            self.test_as_type()
            self.valid = not bool(self.failure_count)
            return

        # test entire implementation
        print(f"Testing entire implementation at {self.base_url}...")
        print("\nMandatory tests:")
        self._log.debug("Testing base info endpoint of %s", BASE_INFO_ENDPOINT)
        base_info = self.test_info_or_links_endpoints(BASE_INFO_ENDPOINT)
        self.get_available_endpoints(base_info)

        for endp in self.test_entry_endpoints:
            entry_info_endpoint = f"{BASE_INFO_ENDPOINT}/{endp}"
            self._log.debug("Testing expected info endpoint %s", entry_info_endpoint)
            self.test_info_or_links_endpoints(entry_info_endpoint)

        for endp in self.test_entry_endpoints:
            self._log.debug("Testing multiple entry endpoint of %s", endp)
            self.test_multi_entry_endpoint(f"{endp}?page_limit={self.page_limit}")

        for endp in self.test_entry_endpoints:
            self._log.debug("Testing single entry request of type %s", endp)
            self.test_single_entry_endpoint(endp)

        for endp in self.endpoint_mandatory_queries:
            # skip empty endpoint query lists
            if self.endpoint_mandatory_queries[endp]:
                self._log.debug("Testing mandatory query syntax on endpoint %s", endp)
                self.test_query_syntax(endp, self.endpoint_mandatory_queries[endp])

        self._log.debug("Testing %s endpoint", LINKS_ENDPOINT)
        self.test_info_or_links_endpoints(LINKS_ENDPOINT)

        self.valid = not bool(self.failure_count)

        print("\nOptional tests:")
        for endp in self.endpoint_optional_queries:
            # skip empty endpoint query lists
            if self.endpoint_mandatory_queries[endp]:
                self._log.debug("Testing optional query syntax on endpoint %s", endp)
                self.test_query_syntax(
                    endp, self.endpoint_optional_queries[endp], optional=True
                )

        if not self.valid:
            print("\n\nFAILURES\n")
            for message in self.failure_messages:
                print_failure(message[0])
                for line in message[1]:
                    print_warning("\t" + line)

        final_message = f"\n\nPassed {self.success_count} out of {self.success_count + self.failure_count} tests."
        if not self.valid:
            print_failure(final_message)
        else:
            print_success(final_message)

        print(
            f"Additionally passed {self.optional_success_count} out of "
            f"{self.optional_success_count + self.optional_failure_count} optional tests."
        )

    def test_info_or_links_endpoints(self, request_str):
        """ Runs the test cases for the info endpoints. """
        response = self.get_endpoint(request_str)
        if response:
            deserialized = self.deserialize_response(
                response, self.response_classes[request_str]
            )
            if not deserialized:
                return response
            return deserialized
        return False

    def test_single_entry_endpoint(self, request_str):
        """ Runs the test cases for the single entry endpoints. """
        _type = request_str.split("?")[0]
        response_cls_name = _type + "/"
        if response_cls_name in self.response_classes:
            response_cls = self.response_classes[response_cls_name]
        else:
            self._log.warning(
                "Deserializing single entry response %s with generic response rather than defined endpoint.",
                _type,
            )
            response_cls = ValidatorEntryResponseOne
        if _type in self.test_id_by_type:
            test_id = self.test_id_by_type[_type]
            response = self.get_endpoint(f"{_type}/{test_id}")
            if response:
                self.deserialize_response(response, response_cls)

    def test_multi_entry_endpoint(self, request_str):
        """ Runs the test cases for the multi entry endpoints. """
        response = self.get_endpoint(request_str)
        _type = request_str.split("?")[0]
        if _type in self.response_classes:
            response_cls = self.response_classes[_type]
        else:
            self._log.warning(
                "Deserializing multi entry response from %s with generic response rather than defined endpoint.",
                _type,
            )
            response_cls = ValidatorEntryResponseMany
        deserialized = self.deserialize_response(response, response_cls)
        self.test_page_limit(response)
        self.get_single_id_from_multi_endpoint(deserialized)

    def test_as_type(self):
        response = self.get_endpoint("")
        if response:
            self._log.debug("Deserialzing response as type %s", self.as_type_cls)
            self.deserialize_response(response, self.as_type_cls)

    @test_case
    def test_page_limit(self, response, check_next_link=True):
        """ Test that a multi-entry endpoint obeys the page limit.

        Parameters:
            response (requests.Response): the response to test for page limit
                compliance.

        Keyword arguments:
            check_next (bool): whether or not to recursively follow and test any
                pagination links provided under `links->next`.

        Raises:
            ResponseError: if test fails in a predictable way.

        Returns:
            bool, str: True if the test was successful, with a string describing
                the success.

        """
        try:
            response = response.json()
        except (AttributeError, json.JSONDecodeError):
            raise ResponseError("Unable to test endpoint page limit.")

        try:
            num_entries = len(response["data"])
        except (KeyError, TypeError):
            raise ResponseError(
                "Response under `data` field was missing or had wrong type."
            )

        if num_entries > self.page_limit:
            raise ResponseError(
                f"Endpoint did not obey page limit: {num_entries} entries vs {self.page_limit} limit"
            )

        try:
            more_data_available = response["meta"]["more_data_available"]
        except KeyError:
            raise ResponseError("Field `meta->more_data_available` was missing.")

        if more_data_available:
            try:
                next_link = response["links"]["next"]
            except KeyError:
                raise ResponseError(
                    "Endpoint suggested more data was available but provided no links->next link."
                )

            next_response = self.get_endpoint(next_link)
            self.test_page_limit(next_response, check_next_link=False)

        return (
            True,
            f"Endpoint obeyed page limit of {self.page_limit} by returning {num_entries} entries.",
        )

    @test_case
    def get_single_id_from_multi_endpoint(self, deserialized):
        """ Scrape an ID from the multi-entry endpoint to use as query
        for single entry endpoint.

        """
        if deserialized and deserialized.data:
            self.test_id_by_type[deserialized.data[0].type] = deserialized.data[0].id
            self._log.debug(
                "Set type %s test ID to %s",
                deserialized.data[0].type,
                deserialized.data[0].id,
            )
        else:
            raise ResponseError("No entries found under endpoint to scrape ID from.")
        return (
            self.test_id_by_type[deserialized.data[0].type],
            f"successfully scraped test ID from {deserialized.data[0].type} endpoint",
        )

    @test_case
    def deserialize_response(self, response: requests.models.Response, response_cls):
        """ Try to create the appropriate pydantic model from the response. """
        if not response:
            raise ResponseError("Request failed")
        return (
            response_cls(**response.json()),
            "deserialized correctly as object of type {}".format(response_cls),
        )

    @test_case
    def get_available_endpoints(self, base_info):
        """ Try to get `entry_types_by_format` even if base info response could not be validated. """
        for _ in [0]:
            available_json_entry_endpoints = []
            try:
                available_json_entry_endpoints = base_info.data.attributes.entry_types_by_format.get(
                    "json"
                )
                break
            except Exception:
                self._log.warning(
                    "Info endpoint failed serialization, trying to manually extract entry_types_by_format."
                )

            if not base_info.json():
                raise ResponseError(
                    "Unable to get entry types from base info endpoint. "
                    f"This may most likely be attributed to a wrong request to the '{BASE_INFO_ENDPOINT}' endpoint."
                )

            try:
                available_json_entry_endpoints = base_info.json()["data"]["attributes"][
                    "entry_types_by_format"
                ]["json"]
                break
            except (KeyError, TypeError):
                raise ResponseError(
                    "Unable to get entry_types_by_format from unserializable base info response {}.".format(
                        base_info
                    )
                )
        else:
            raise ResponseError(
                "Unable to find any JSON entry types in entry_types_by_format"
            )

        if self.index and available_json_entry_endpoints != []:
            raise ResponseError(
                "No entry endpoint are allowed for an Index meta-database"
            )

        self.test_entry_endpoints |= set(available_json_entry_endpoints)
        for non_entry_endpoint in ("info", "links"):
            if non_entry_endpoint in self.test_entry_endpoints:
                raise ResponseError(
                    f'Illegal entry "{non_entry_endpoint}" was found in entry_types_by_format"'
                )
        return (
            available_json_entry_endpoints,
            "successfully found available entry types in baseinfo",
        )

    @test_case
    def get_endpoint(self, request_str, optional=False):
        """ Gets the response from the endpoint specified by `request_str`. """

        request_str = request_str.replace("\n", "")
        response = self.client.get(request_str)

        if response.status_code != 200:
            error_titles = [
                error.get("title", "") for error in response.json().get("errors", [])
            ]
            error_details = [
                error.get("detail", "") for error in response.json().get("errors", [])
            ]
            message = f"Request to '{request_str}' returned HTTP code {response.status_code}. Errors:\n"
            for error_title, error_detail in zip(error_titles, error_details):
                message += f"{error_title}: {error_details}"
            raise ResponseError(message)

        return response, "request successful."

    def test_query_syntax(self, endpoint, endpoint_queries, optional=False):
        """ Execute a list of valid queries agains the endpoint and assert
        that no errors are raised.

        Parameters:
            endpoint (str): the endpoint to query (e.g. "structures").
            endpoint_queries (list): the list of valid mandatory queries
                for that endpoint, where the queries do not include the
                "?filter=" prefix, e.g. ['elements HAS "Na"'].

        Keyword arguments:
            optional (bool): treat the success of the queries as optional.

        """

        valid_queries = [f"{endpoint}?filter={query}" for query in endpoint_queries]
        for query in valid_queries:
            self.get_endpoint(query, optional=optional)



# pylint: disable=relative-beyond-top-level,import-outside-toplevel
import unittest
import mongomock

from optimade.server.mappers import BaseResourceMapper
from optimade.server.entry_collections import MongoCollection
from optimade.models import StructureResource


class ResourceMapperTests(unittest.TestCase):
    def test_disallowed_aliases(self):
        class MyMapper(BaseResourceMapper):
            ALIASES = (("$and", "my_special_and"), ("not", "$not"))

        mapper = MyMapper()
        toy_collection = mongomock.MongoClient()["fake"]["fake"]
        with self.assertRaises(RuntimeError):
            MongoCollection(toy_collection, StructureResource, mapper)

    def test_property_aliases(self):
        class MyMapper(BaseResourceMapper):
            PROVIDER_FIELDS = ("dft_parameters", "test_field")
            LENGTH_ALIASES = (("_exmpl_test_field", "test_field_len"),)
            ALIASES = (("field", "completely_different_field"),)

        mapper = MyMapper()
        self.assertEqual(mapper.alias_for("_exmpl_dft_parameters"), "dft_parameters")
        self.assertEqual(mapper.alias_for("_exmpl_test_field"), "test_field")
        self.assertEqual(mapper.alias_for("field"), "completely_different_field")
        self.assertEqual(mapper.length_alias_for("_exmpl_test_field"), "test_field_len")
        self.assertEqual(mapper.length_alias_for("test_field"), None)
        self.assertEqual(mapper.alias_for("test_field"), "test_field")

        # nested properties
        self.assertEqual(
            mapper.alias_for("_exmpl_dft_parameters.nested.property"),
            "dft_parameters.nested.property",
        )
        self.assertEqual(
            mapper.alias_for("_exmpl_dft_parameters.nested_property"),
            "dft_parameters.nested_property",
        )

        # test nonsensical query
        self.assertEqual(mapper.alias_for("_exmpl_test_field."), "test_field.")

        # test an awkward case that has no alias
        self.assertEqual(
            mapper.alias_for("_exmpl_dft_parameters_dft_parameters.nested.property"),
            "_exmpl_dft_parameters_dft_parameters.nested.property",
        )



import numpy as np
from ase.io import read

from nomadcore.atoms2nomad import ase_atoms_to_section_system

class CubeError(OSError):
    pass


def read_cube_file(backend, file_name):
    try:
        d = read(file_name, format = 'cube', full_output = True)

    except Exception as err:
        raise CubeError(err)

    data = d['data']
    atoms = d['atoms']
    origin = d['origin']
    nx, ny, nz = data.shape
    displacements = np.array([atoms.cell[i]/data.shape[i] for i in range(3)])

    system = ase_atoms_to_section_system(backend, atoms)

    singleconfig = backend.openSection('section_single_configuration_calculation')
    volumetric = backend.openSection('section_volumetric_data')

    backend.addValue('volumetric_data_nx', nx)
    backend.addValue('volumetric_data_ny', ny)
    backend.addValue('volumetric_data_nz', nz)

    backend.addArrayValues('volumetric_data_origin', origin)
    backend.addArrayValues('volumetric_data_displacements', displacements)

    backend.addValue('volumetric_data_multiplicity', 1)
    backend.addArrayValues('volumetric_data_values', data[None])
    backend.closeSection('section_volumetric_data', volumetric)

    backend.addValue('single_configuration_calculation_to_system_ref', system)
    backend.closeSection('section_single_configuration_calculation', singleconfig)



import os
import numpy as np
import logging

from nomadcore.unit_conversion import unit_conversion
from nomad.datamodel.metainfo.public import section_run, section_system,\
    section_single_configuration_calculation, section_method, section_calculation_to_calculation_refs,\
    Workflow, Elastic

from elasticparser.metainfo.elastic import x_elastic_section_strain_diagrams,\
    x_elastic_section_fitting_parameters

from elasticparser.elastic_properties import ElasticProperties

a_to_m = unit_conversion.convert_unit_function('angstrom', 'meter')
ha_to_J = unit_conversion.convert_unit(1, 'hartree', 'J')
giga = 10 ** 9


class ElasticParserInterface:
    def __init__(self, filepath, archive, logger=None):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.logger = logger if logger is not None else logging
        self.properties = ElasticProperties(self.filepath)

    def parse_strain(self):
        sec_scc = self.archive.section_run[-1].section_single_configuration_calculation[-1]
        method = self.properties.info['calculation_method'].lower()

        if method == 'energy':
            strain, energy = self.properties.get_strain_energy()
            if not strain:
                self.logger.warn('Error getting strain and energy data')
                return

            n_strains = self.properties.info['n_strains']

            sec_strain_diagram = sec_scc.m_create(x_elastic_section_strain_diagrams)
            sec_strain_diagram.x_elastic_strain_diagram_type = 'energy'
            sec_strain_diagram.x_elastic_strain_diagram_number_of_eta = len(strain[0])
            sec_strain_diagram.x_elastic_strain_diagram_eta_values = strain
            sec_strain_diagram.x_elastic_strain_diagram_values = np.array(energy) * ha_to_J

            poly_fit_2 = int((n_strains - 1) / 2)

            poly_fit = {
                '2nd': poly_fit_2, '3rd': poly_fit_2 - 1, '4th': poly_fit_2 - 1,
                '5th': poly_fit_2 - 2, '6th': poly_fit_2 - 2, '7th': poly_fit_2 - 3}

            energy_fit = self.properties.get_energy_fit()
            if not energy_fit:
                self.logger.warn('Error getting energy fit data')
                return

            for diagram_type in ['cross-validation', 'd2e']:
                for fit_order in energy_fit[diagram_type][0].keys():
                    sec_strain_diagram = sec_scc.m_create(x_elastic_section_strain_diagrams)
                    sec_strain_diagram.x_elastic_strain_diagram_type = diagram_type
                    sec_strain_diagram.x_elastic_strain_diagram_polynomial_fit_order = int(fit_order[:-2])
                    sec_strain_diagram.x_elastic_strain_diagram_number_of_eta = poly_fit.get(fit_order, None)
                    sec_strain_diagram.x_elastic_strain_diagram_eta_values = energy_fit[diagram_type][0][fit_order]
                    convert = giga if diagram_type == 'd2e' else ha_to_J
                    sec_strain_diagram.x_elastic_strain_diagram_values = np.array(energy_fit[diagram_type][1][fit_order]) * convert

        elif method == 'stress':
            strain, stress = self.properties.get_strain_stress()
            for diagram_type in ['Lagrangian-stress', 'Physical-stress']:
                strain_i = strain[diagram_type]
                stress_i = np.transpose(np.array(stress[diagram_type]), axes=(2, 0, 1))
                if not strain_i:
                    continue

                for si in range(6):
                    sec_strain_diagram = sec_scc.m_create(x_elastic_section_strain_diagrams)
                    sec_strain_diagram.x_elastic_strain_diagram_type = diagram_type
                    sec_strain_diagram.x_elastic_strain_diagram_stress_Voigt_component = si + 1
                    sec_strain_diagram.x_elastic_strain_diagram_number_of_eta = len(strain_i[0])
                    sec_strain_diagram.x_elastic_strain_diagram_eta_values = strain_i
                    sec_strain_diagram.x_elastic_strain_diagram_values = stress_i[si]

            stress_fit = self.properties.get_stress_fit()
            for diagram_type in ['cross-validation', 'dtn']:
                if stress_fit.get(diagram_type, None) is None:
                    continue

                for si in range(6):
                    for fit_order in stress_fit[diagram_type][si][0].keys():
                        sec_strain_diagram = sec_scc.m_create(x_elastic_section_strain_diagrams)
                        sec_strain_diagram.x_elastic_strain_diagram_type = diagram_type
                        sec_strain_diagram.x_elastic_strain_diagram_stress_Voigt_component = si + 1
                        sec_strain_diagram.x_elastic_strain_diagram_polynomial_fit_order = int(fit_order[:-2])
                        sec_strain_diagram.x_elastic_strain_diagram_number_of_eta = poly_fit.get(fit_order, None)
                        sec_strain_diagram.x_elastic_strain_diagram_eta_values = stress_fit[diagram_type][si][0][fit_order]
                        convert = giga if diagram_type == 'dtn' else ha_to_J
                        sec_strain_diagram.x_elastic_strain_diagram_values = np.array(stress_fit[diagram_type][si][1][fit_order]) * convert

    def parse_elastic_constant(self):
        sec_scc = self.archive.section_run[-1].section_single_configuration_calculation[-1]

        order = self.properties.info['order']

        if order == 2:
            matrices, moduli, eigenvalues = self.properties.get_elastic_constants_order2()

            sec_scc.x_elastic_2nd_order_constants_notation_matrix = matrices['voigt']
            sec_scc.x_elastic_2nd_order_constants_matrix = matrices['elastic_constant'] * giga
            sec_scc.x_elastic_2nd_order_constants_compliance_matrix = matrices['compliance'] / giga

            sec_scc.x_elastic_Voigt_bulk_modulus = moduli.get('B_V', moduli.get('K_V')) * giga
            sec_scc.x_elastic_Voigt_shear_modulus = moduli['G_V'] * giga

            sec_scc.x_elastic_Reuss_bulk_modulus = moduli.get('B_R', moduli.get('K_R')) * giga
            sec_scc.x_elastic_Reuss_shear_modulus = moduli['G_R'] * giga

            sec_scc.x_elastic_Hill_bulk_modulus = moduli.get('B_H', moduli.get('K_H')) * giga
            sec_scc.x_elastic_Hill_shear_modulus = moduli['G_H'] * giga

            sec_scc.x_elastic_Voigt_Young_modulus = moduli['E_V'] * giga
            sec_scc.x_elastic_Voigt_Poisson_ratio = moduli['nu_V']
            sec_scc.x_elastic_Reuss_Young_modulus = moduli['E_R'] * giga
            sec_scc.x_elastic_Reuss_Poisson_ratio = moduli['nu_R']
            sec_scc.x_elastic_Hill_Young_modulus = moduli['E_H'] * giga
            sec_scc.x_elastic_Hill_Poisson_ratio = moduli['nu_H']

            sec_scc.x_elastic_eigenvalues = eigenvalues * giga

        elif order == 3:
            elastic_constant = self.properties.get_elastic_constants_order3()

            sec_scc.x_elastic_3rd_order_constants_matrix = elastic_constant

    def parse(self):

        sec_run = self.archive.m_create(section_run)

        sec_run.program_name = 'elastic'
        sec_run.program_version = '1.0'

        sec_system = sec_run.m_create(section_system)

        symbols, positions, cell = self.properties.get_structure_info()
        volume = self.properties.info['equilibrium_volume']
        volume = a_to_m(a_to_m(a_to_m(volume)))

        sec_system.atom_labels = symbols
        sec_system.atom_positions = a_to_m(positions)
        sec_system.simulation_cell = a_to_m(cell)
        sec_system.configuration_periodic_dimensions = [True, True, True]
        sec_system.x_elastic_space_group_number = self.properties.info['space_group_number']
        sec_system.x_elastic_unit_cell_volume = volume

        sec_method = sec_run.m_create(section_method)
        sec_method.x_elastic_elastic_constant_order = self.properties.info['order']
        sec_method.x_elastic_calculation_method = self.properties.info['calculation_method']
        sec_method.x_elastic_code = self.properties.info['code_name']
        sec_method.x_elastic_max_lagrangian_strain = self.properties.info['max_strain']
        sec_method.x_elastic_number_of_distorted_structures = self.properties.info['n_strains']

        deformation_types = self.properties.get_deformation_types()
        sec_method.x_elastic_deformation_types = deformation_types
        sec_method.x_elastic_number_of_deformations = len(self.properties.deformation_dirs)

        references = self.properties.get_references_to_calculations()
        sec_scc = sec_run.m_create(section_single_configuration_calculation)
        for reference in references:
            sec_calc_ref = sec_scc.m_create(section_calculation_to_calculation_refs)
            sec_calc_ref.calculation_to_calculation_external_url = reference
            sec_calc_ref.calculation_to_calculation_kind = 'source_calculation'

        fit_input = self.properties.get_input()
        sec_fit_par = sec_method.m_create(x_elastic_section_fitting_parameters)
        sec_fit_par.x_elastic_fitting_parameters_eta = fit_input[0]
        sec_fit_par.x_elastic_fitting_parameters_polynomial_order = fit_input[1]

        self.parse_strain()

        self.parse_elastic_constant()

        sec_scc.single_configuration_to_calculation_method_ref = sec_method
        sec_scc.single_configuration_calculation_to_system_ref = sec_system

        sec_workflow = self.archive.m_create(Workflow)
        sec_workflow.workflow_type = 'elastic'
        sec_elastic = sec_workflow.m_create(Elastic)
        sec_elastic.elastic_calculation_method = self.properties.info['calculation_method'].lower()
        sec_elastic.elastic_constants_order = self.properties.info['order']
        sec_elastic.strain_maximum = self.properties.info['max_strain']



import logging
LOGGER = logging.getLogger("nomad")


"""List of atomic symbols in order."""
_symbols = [
    '',   'H',  'He', 'Li', 'Be',
    'B',  'C',  'N',  'O',  'F',
    'Ne', 'Na', 'Mg', 'Al', 'Si',
    'P',  'S',  'Cl', 'Ar', 'K',
    'Ca', 'Sc', 'Ti', 'V',  'Cr',
    'Mn', 'Fe', 'Co', 'Ni', 'Cu',
    'Zn', 'Ga', 'Ge', 'As', 'Se',
    'Br', 'Kr', 'Rb', 'Sr', 'Y',
    'Zr', 'Nb', 'Mo', 'Tc', 'Ru',
    'Rh', 'Pd', 'Ag', 'Cd', 'In',
    'Sn', 'Sb', 'Te', 'I',  'Xe',
    'Cs', 'Ba', 'La', 'Ce', 'Pr',
    'Nd', 'Pm', 'Sm', 'Eu', 'Gd',
    'Tb', 'Dy', 'Ho', 'Er', 'Tm',
    'Yb', 'Lu', 'Hf', 'Ta', 'W',
    'Re', 'Os', 'Ir', 'Pt', 'Au',
    'Hg', 'Tl', 'Pb', 'Bi', 'Po',
    'At', 'Rn', 'Fr', 'Ra', 'Ac',
    'Th', 'Pa', 'U',  'Np', 'Pu',
    'Am', 'Cm', 'Bk', 'Cf', 'Es',
    'Fm', 'Md', 'No', 'Lr'
]

"""Dict of atomic symbols with their atomic number"""
_numbers = {
    'H': 1,  'He': 2, 'Li': 3, 'Be': 4,
    'B': 5,  'C': 6,  'N': 7,  'O': 8,  'F': 9,
    'Ne': 10, 'Na': 11, 'Mg': 12, 'Al': 13, 'Si': 14,
    'P': 15,  'S': 16,  'Cl': 17, 'Ar': 18, 'K': 19,
    'Ca': 20, 'Sc': 21, 'Ti': 22, 'V': 23,  'Cr': 24,
    'Mn': 25, 'Fe': 26, 'Co': 27, 'Ni': 28, 'Cu': 29,
    'Zn': 30, 'Ga': 31, 'Ge': 32, 'As': 33, 'Se': 34,
    'Br': 35, 'Kr': 36, 'Rb': 37, 'Sr': 38, 'Y': 39,
    'Zr': 40, 'Nb': 41, 'Mo': 42, 'Tc': 43, 'Ru': 44,
    'Rh': 45, 'Pd': 46, 'Ag': 47, 'Cd': 48, 'In': 49,
    'Sn': 50, 'Sb': 51, 'Te': 52, 'I': 53,  'Xe': 54,
    'Cs': 55, 'Ba': 56, 'La': 57, 'Ce': 58, 'Pr': 59,
    'Nd': 60, 'Pm': 61, 'Sm': 62, 'Eu': 63, 'Gd': 64,
    'Tb': 65, 'Dy': 66, 'Ho': 67, 'Er': 68, 'Tm': 69,
    'Yb': 70, 'Lu': 71, 'Hf': 72, 'Ta': 73, 'W': 74,
    'Re': 75, 'Os': 76, 'Ir': 77, 'Pt': 78, 'Au': 79,
    'Hg': 80, 'Tl': 81, 'Pb': 82, 'Bi': 83, 'Po': 84,
    'At': 85, 'Rn': 86, 'Fr': 87, 'Ra': 88, 'Ac': 89,
    'Th': 90, 'Pa': 91, 'U': 92,  'Np': 93, 'Pu': 94,
    'Am': 95, 'Cm': 96, 'Bk': 97, 'Cf': 98, 'Es': 99,
    'Fm': 100, 'Md': 101, 'No': 102, 'Lr': 103,
}


def get_atom_number(symbol):
    """Given an element symbol, return the atomic number (number of protons).

    Args:
        symbol(str): The atomic symbol for the given element.

    Returns:
        The atomic number (number of protons) of the element. If no species
        could be found, returns None.
    """
    number = _numbers.get(symbol)
    if number is None:
        raise KeyError("The element symbol '{}' is not valid. Remember to use the capitalized form of the name.".format(symbol))
    return number


def get_atom_symbol(number):
    """Given an atomic number, returns the symbol for that element.

    Args:
        number(int): The atomic number (number of protons) of the element.

    Returns:
        The atomic symbol for the given element. If no species could be found,
        returns None.
    """
    try:
        symbol = _symbols[number]
    except IndexError:
        raise IndexError("No element with atomic number '{}'.".format(number))

    return symbol
